import time
from datetime import datetime, timedelta
from typing import List, Optional

import alpaca_trade_api as tradeapi
import requests
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog

from .base import Scanner


class Momentum(Scanner):
    name = "momentum"

    def __init__(
        self,
        provider: str,
        recurrence: Optional[timedelta],
        data_api: tradeapi,
        max_share_price: float,
        min_share_price: float,
        min_last_dv: float,
        today_change_percent: float,
        min_volume: float,
        from_market_open: float,
        max_symbols: int = config.total_tickers,
    ):
        self.provider = provider
        self.max_share_price = max_share_price
        self.min_share_price = min_share_price
        self.min_last_dv = min_last_dv
        self.min_volume = min_volume
        self.today_change_percent = today_change_percent
        self.from_market_open = from_market_open
        self.max_symbols = max_symbols

        super().__init__(
            name=self.name, recurrence=recurrence, data_api=data_api,
        )

    @classmethod
    def __str__(cls) -> str:
        return cls.name

    def _wait_time(self) -> None:
        if config.market_open:
            nyc = timezone("America/New_York")
            since_market_open = (
                datetime.today().astimezone(nyc) - config.market_open
            )

            if since_market_open.seconds // 60 < self.from_market_open:
                tlog(f"market open, wait {self.from_market_open} minutes")
                while since_market_open.seconds // 60 < self.from_market_open:
                    time.sleep(1)
                    since_market_open = (
                        datetime.today().astimezone(nyc) - config.market_open
                    )

        tlog(f"Scanner {self.name} ready to run")

    def _get_trade_able_symbols(self) -> List[str]:
        assets = self.data_api.list_assets()
        tlog(f"loaded list of {len(assets)} trade-able assets from Alpaca")

        trade_able_symbols = [
            asset.symbol for asset in assets if asset.tradable
        ]
        tlog(
            f"total number of trade-able symbols is {len(trade_able_symbols)}"
        )
        return trade_able_symbols

    def run_polygon(self) -> List[str]:
        tlog(f"{self.name}: run_polygon(): started")
        try:
            while True:
                tickers = self.data_api.polygon.all_tickers()
                tlog(f"loaded {len(tickers)} tickers from Polygon")
                if not len(tickers):
                    break
                trade_able_symbols = self._get_trade_able_symbols()

                unsorted = [
                    ticker
                    for ticker in tickers
                    if (
                        ticker.ticker in trade_able_symbols
                        and self.max_share_price
                        >= ticker.lastTrade["p"]
                        >= self.min_share_price
                        and ticker.prevDay["v"] * ticker.lastTrade["p"]
                        > self.min_last_dv
                        and ticker.todaysChangePerc
                        >= self.today_change_percent
                        and (
                            ticker.day["v"] > self.min_share_price
                            or config.bypass_market_schedule
                        )
                    )
                ]
                if len(unsorted) > 0:
                    ticker_by_volume = sorted(
                        unsorted,
                        key=lambda ticker: float(ticker.day["v"]),
                        reverse=True,
                    )
                    tlog(f"picked {len(ticker_by_volume)} symbols")
                    return [x.ticker for x in ticker_by_volume][
                        : self.max_symbols
                    ]

                tlog("did not find gaping stock, retrying")
                time.sleep(30)
        except KeyboardInterrupt:
            tlog("KeyboardInterrupt")
            pass

        return []

    def run_finnhub(self) -> List[str]:
        tlog(f"{self.name}: run_finnhub(): started")
        trade_able_symbols = self._get_trade_able_symbols()

        nyc = timezone("America/New_York")
        _from = datetime.today().astimezone(nyc) - timedelta(days=1)
        _to = datetime.now(nyc)
        symbols = []
        try:
            with requests.Session() as s:
                for symbol in trade_able_symbols:
                    try:
                        retry = True
                        while retry:
                            retry = False
                            url = (
                                f"{config.finnhub_base_url}/stock/candle?symbol={symbol}&"
                                f"resolution=D&from={_from.strftime('%s')}&to={_to.strftime('%s')}&"
                                f"token={config.finnhub_api_key}"
                            )

                            try:
                                r = s.get(url)
                            except ConnectionError:
                                retry = True
                                continue

                            if r.status_code == 200:
                                response = r.json()
                            if response["s"] != "no_data":
                                prev_prec = (
                                    100.0
                                    * (response["c"][1] - response["c"][0])
                                    / response["c"][0]
                                )
                                if (
                                    self.max_share_price
                                    > response["c"][1]
                                    > self.min_share_price
                                    and response["v"][0] * response["c"][0]
                                    > self.min_last_dv
                                    and prev_prec > self.today_change_percent
                                    and response["v"][1] > self.min_volume
                                ):
                                    symbols.append(symbol)
                                    tlog(
                                        f"collected {len(symbols)}/{self.max_symbols}"
                                    )
                                    if len(symbols) == self.max_symbols:
                                        break

                            elif r.status_code == 429:
                                tlog(
                                    f"{trade_able_symbols.index(symbol)}/{len(trade_able_symbols)} API limit: ({r.text})"
                                )
                                time.sleep(30)
                                retry = True
                            else:
                                print(r.status_code, r.text)

                    except IndexError:
                        pass

        except KeyboardInterrupt:
            tlog("KeyboardInterrupt")
            pass

        tlog(f"loaded {len(symbols)} from Finnhub")
        return symbols

    def run(self) -> List[str]:
        self._wait_time()

        if self.provider == "polygon":
            return self.run_polygon()
        elif self.provider == "finnhub":
            return self.run_finnhub()
        else:
            raise Exception(
                f"Invalid provider {self.provider} for scanner {self.name}"
            )



from datetime import datetime, timedelta
from typing import Dict, List, Tuple

import alpaca_trade_api as tradeapi
import numpy as np
import talib
from pandas import DataFrame as df
from talib import BBANDS, MACD, RSI

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog
from liualgotrader.common.trading_data import (buy_indicators, buy_time,
                                               cool_down, last_used_strategy,
                                               latest_cost_basis,
                                               latest_scalp_basis, open_orders,
                                               sell_indicators, stop_prices,
                                               target_prices)
from liualgotrader.fincalcs.support_resistance import find_stop
from liualgotrader.strategies.base import Strategy, StrategyType


class MomentumLong(Strategy):
    name = "momentum_long"

    def __init__(
        self,
        batch_id: str,
        schedule: List[Dict],
        ref_run_id: int = None,
        check_patterns: bool = False,
    ):
        self.check_patterns = check_patterns
        super().__init__(
            name=self.name,
            type=StrategyType.DAY_TRADE,
            batch_id=batch_id,
            ref_run_id=ref_run_id,
            schedule=schedule,
        )

    async def buy_callback(self, symbol: str, price: float, qty: int) -> None:
        latest_scalp_basis[symbol] = latest_cost_basis[symbol] = price

    async def sell_callback(self, symbol: str, price: float, qty: int) -> None:
        latest_scalp_basis[symbol] = price

    async def create(self) -> None:
        await super().create()
        tlog(f"strategy {self.name} created")

    async def should_cool_down(self, symbol: str, now: datetime):
        if (
            symbol in cool_down
            and cool_down[symbol]
            and cool_down[symbol] >= now.replace(second=0, microsecond=0)  # type: ignore
        ):
            return True

        cool_down[symbol] = None
        return False

    async def run(
        self,
        symbol: str,
        position: int,
        minute_history: df,
        now: datetime,
        portfolio_value: float = None,
        trading_api: tradeapi = None,
        debug: bool = False,
        backtesting: bool = False,
    ) -> Tuple[bool, Dict]:
        data = minute_history.iloc[-1]
        prev_min = minute_history.iloc[-2]

        morning_rush = (
            True if (now - config.market_open).seconds // 60 < 30 else False
        )

        if (
            await super().is_buy_time(now)
            and not position
            and not await self.should_cool_down(symbol, now)
        ):
            # Check for buy signals
            lbound = config.market_open
            ubound = lbound + timedelta(minutes=15)

            if debug:
                tlog(f"15 schedule {lbound}/{ubound}")
            try:
                high_15m = minute_history[lbound:ubound][  # type: ignore
                    "high"
                ].max()
            except Exception as e:
                tlog(
                    f"[{self.name}] error aggregation {e} - maybe should use nearest?"
                )
                return False, {}

            if data.close > high_15m or (
                hasattr(config, "bypass_market_schedule")
                and config.bypass_market_schedule
            ):
                close = (
                    minute_history["close"]
                    .dropna()
                    .between_time("9:30", "16:00")
                )
                close_5m = (
                    minute_history["close"]
                    .dropna()
                    .between_time("9:30", "16:00")
                    .resample("5min")
                    .last()
                ).dropna()

                macds = MACD(close)
                # sell_macds = MACD(close, 13, 21)

                macd = macds[0]
                macd_signal = macds[1]
                macd_hist = macds[2]
                macd_trending = macd[-3] < macd[-2] < macd[-1]
                macd_above_signal = macd[-1] > macd_signal[-1] * 1.1
                macd_hist_trending = (
                    macd_hist[-3] < macd_hist[-2] < macd_hist[-1]
                )

                if (
                    macd[-1] > 0
                    and macd_trending
                    and macd_above_signal
                    and macd_hist_trending
                ):
                    macd2 = MACD(close, 40, 60)[0]
                    if macd2[-1] >= 0 and np.diff(macd2)[-1] >= 0:
                        if debug:
                            tlog(
                                f"[{self.name}][{now}] slow macd confirmed trend"
                            )

                        # check RSI does not indicate overbought
                        rsi = RSI(close, 14)

                        if debug:
                            tlog(
                                f"[{self.name}][{now}] {symbol} RSI={round(rsi[-1], 2)}"
                            )

                        rsi_limit = 75
                        if rsi[-1] < rsi_limit:
                            if debug:
                                tlog(
                                    f"[{self.name}][{now}] {symbol} RSI {round(rsi[-1], 2)} <= {rsi_limit}"
                                )
                        else:
                            tlog(
                                f"[{self.name}][{now}] {symbol} RSI over-bought, cool down for 5 min"
                            )
                            cool_down[symbol] = now.replace(
                                second=0, microsecond=0
                            ) + timedelta(minutes=5)

                            return False, {}

                        stop_price = find_stop(
                            data.close if not data.vwap else data.vwap,
                            minute_history,
                            now,
                        )
                        target_price = (
                            3 * (data.close - stop_price) + data.close
                        )
                        target_prices[symbol] = target_price
                        stop_prices[symbol] = stop_price

                        if portfolio_value is None:
                            if trading_api:
                                portfolio_value = float(
                                    trading_api.get_account().portfolio_value
                                )
                            else:
                                raise Exception(
                                    f"{self.name}: both portfolio_value and trading_api can't be None"
                                )

                        shares_to_buy = (
                            portfolio_value
                            * config.risk
                            // (data.close - stop_prices[symbol])
                        )
                        if not shares_to_buy:
                            shares_to_buy = 1
                        shares_to_buy -= position
                        if shares_to_buy > 0:
                            buy_price = max(data.close, data.vwap)
                            tlog(
                                f"[{self.name}][{now}] Submitting buy for {shares_to_buy} shares of {symbol} at {buy_price} target {target_prices[symbol]} stop {stop_prices[symbol]}"
                            )

                            # await asyncio.sleep(0)
                            buy_indicators[symbol] = {
                                "macd": macd[-5:].tolist(),
                                "macd_signal": macd_signal[-5:].tolist(),
                                "vwap": data.vwap,
                                "avg": data.average,
                            }

                            return (
                                True,
                                {
                                    "side": "buy",
                                    "qty": str(shares_to_buy),
                                    "type": "limit",
                                    "limit_price": str(buy_price),
                                }
                                if not morning_rush
                                else {
                                    "side": "buy",
                                    "qty": str(shares_to_buy),
                                    "type": "market",
                                },
                            )
            else:
                if debug:
                    tlog(f"[{self.name}][{now}] {data.close} < 15min high ")
        if (
            await super().is_sell_time(now)
            and position > 0
            and symbol in latest_cost_basis
            and last_used_strategy[symbol].name == self.name
        ):
            if open_orders.get(symbol) is not None:
                tlog(
                    f"momentum_long: open order for {symbol} exists, skipping"
                )
                return False, {}

            macds = MACD(
                minute_history["close"].dropna().between_time("9:30", "16:00"),
                13,
                21,
            )
            macd = macds[0]
            macd_signal = macds[1]
            rsi = RSI(
                minute_history["close"].dropna().between_time("9:30", "16:00"),
                14,
            )

            movement = (
                data.close - latest_scalp_basis[symbol]
            ) / latest_scalp_basis[symbol]
            macd_val = macd[-1]
            macd_signal_val = macd_signal[-1]

            round_factor = (
                2 if macd_val >= 0.1 or macd_signal_val >= 0.1 else 3
            )
            scalp_threshold = (
                target_prices[symbol] + latest_scalp_basis[symbol]
            ) / 2.0

            macd_below_signal = round(macd_val, round_factor) < round(
                macd_signal_val, round_factor
            )
            bail_out = (
                latest_scalp_basis[symbol] > latest_cost_basis[symbol]
                and macd_below_signal
                and round(macd[-1], round_factor)
                < round(macd[-2], round_factor)
            )

            scalp = movement > 0.02 or data.vwap > scalp_threshold
            below_cost_base = data.vwap < latest_cost_basis[symbol]

            rsi_limit = 79 if not morning_rush else 85
            to_sell = False
            partial_sell = False
            limit_sell = False
            sell_reasons = []
            if data.close <= stop_prices[symbol]:
                to_sell = True
                sell_reasons.append("stopped")
            elif (
                below_cost_base
                and round(macd_val, 2) < 0
                and rsi[-1] < rsi[-2]
                and round(macd[-1], round_factor)
                < round(macd[-2], round_factor)
                and data.vwap < 0.95 * data.average
            ):
                to_sell = True
                sell_reasons.append(
                    "below cost & macd negative & RSI trending down and too far from VWAP"
                )
            elif data.close >= target_prices[symbol] and macd[-1] <= 0:
                to_sell = True
                sell_reasons.append("above target & macd negative")
            elif rsi[-1] >= rsi_limit:
                to_sell = True
                sell_reasons.append("rsi max, cool-down for 5 minutes")
                cool_down[symbol] = now.replace(
                    second=0, microsecond=0
                ) + timedelta(minutes=5)
            elif bail_out:
                to_sell = True
                sell_reasons.append("bail")
            elif scalp:
                partial_sell = True
                to_sell = True
                sell_reasons.append("scale-out")

            if to_sell:
                sell_indicators[symbol] = {
                    "rsi": rsi[-3:].tolist(),
                    "movement": movement,
                    "sell_macd": macd[-5:].tolist(),
                    "sell_macd_signal": macd_signal[-5:].tolist(),
                    "vwap": data.vwap,
                    "avg": data.average,
                    "reasons": " AND ".join(
                        [str(elem) for elem in sell_reasons]
                    ),
                }

                if not partial_sell:
                    if not limit_sell:
                        tlog(
                            f"[{self.name}][{now}] Submitting sell for {position} shares of {symbol} at market with reason:{sell_reasons}"
                        )
                        return (
                            True,
                            {
                                "side": "sell",
                                "qty": str(position),
                                "type": "market",
                            },
                        )
                    else:
                        tlog(
                            f"[{self.name}][{now}] Submitting sell for {position} shares of {symbol} at {data.close} with reason:{sell_reasons}"
                        )
                        return (
                            True,
                            {
                                "side": "sell",
                                "qty": str(position),
                                "type": "limit",
                                "limit_price": str(data.close),
                            },
                        )
                else:
                    qty = int(position / 2) if position > 1 else 1
                    tlog(
                        f"[{self.name}][{now}] Submitting sell for {str(qty)} shares of {symbol} at limit of {data.close }with reason:{sell_reasons}"
                    )
                    return (
                        True,
                        {
                            "side": "sell",
                            "qty": str(qty),
                            "type": "limit",
                            "limit_price": str(data.close),
                        },
                    )

        return False, {}



import json
from datetime import datetime, timedelta
from typing import Any, Dict, List, Tuple

from asyncpg.pool import Pool

from liualgotrader.common import config


class AlgoRun:
    def __init__(self, strategy_name: str, batch_id: str):
        self.run_id: Any[None, int] = None
        self.strategy_name = strategy_name
        self.batch_id = batch_id

    async def save(
        self, pool: Pool = None, env: str = None, ref_algo_run_id: int = None
    ) -> None:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                if not ref_algo_run_id:
                    q = """
                        INSERT INTO algo_run (algo_name, algo_env, build_number, parameters, batch_id)
                        VALUES ($1, $2, $3, $4, $5)
                        RETURNING algo_run_id
                        """

                    self.run_id = await con.fetchval(
                        q,
                        self.strategy_name,
                        env if env else config.env,
                        config.build_label,
                        json.dumps({"DSN": config.dsn,}),
                        self.batch_id,
                    )
                else:
                    q = """
                        INSERT INTO algo_run (algo_name, algo_env, build_number, parameters, batch_id, ref_algo_run)
                        VALUES ($1, $2, $3, $4, $5, $6)
                        RETURNING algo_run_id
                        """
                    self.run_id = await con.fetchval(
                        q,
                        self.strategy_name,
                        env if env else config.env,
                        config.build_label,
                        json.dumps({"DSN": config.dsn,}),
                        self.batch_id,
                        ref_algo_run_id,
                    )

    async def update_end_time(self, pool: Pool, end_reason: str):
        async with pool.acquire() as con:
            async with con.transaction():
                await con.execute(
                    """
                        UPDATE algo_run SET end_time='now()',end_reason=$1
                        WHERE algo_run_id = $2
                    """,
                    end_reason,
                    self.run_id,
                )

    @classmethod
    async def get_batches(
        cls, pool: Pool = None, nax_batches: int = 30
    ) -> Dict[str, List[str]]:
        rc: Dict = {}
        if not pool:
            pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT batch_id, algo_run_id, algo_name, algo_env, build_number, start_time
                        FROM algo_run
                        ORDER BY start_time DESC
                        LIMIT $1
                    """,
                    nax_batches,
                )

                if rows:
                    for row in rows:
                        if row[0] not in rc:
                            rc[row[0]] = [list(row.values())[1:]]
                        else:
                            rc[row[0]].append(list(row.values())[1:])

        return rc

    @classmethod
    async def get_batch_details(
        cls, batch_id: str, pool: Pool = None
    ) -> List[Tuple[int, datetime, str]]:
        rc: List = []
        if not pool:
            pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT algo_run_id, start_time, parameters, algo_name
                        FROM algo_run
                        WHERE batch_id = $1
                        ORDER BY start_time DESC
                    """,
                    batch_id,
                )

                if rows:
                    rc = [(row[0], row[1], row[3],) for row in rows]

        return rc



from pandas import DataFrame as df
from pandas import Timestamp as ts
from tabulate import tabulate

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


def add_daily_vwap(minute_data: df, debug=False) -> bool:
    back_time = ts(config.market_open)

    try:
        back_time_index = minute_data["close"].index.get_loc(
            back_time, method="nearest"
        )
    except IndexError as e:
        if debug:
            tlog(
                f"IndexError exception {e} in add_daily_vwap for {minute_data}"
            )
        return False

    minute_data["pv"] = minute_data.apply(
        lambda x: (x["close"] + x["high"] + x["low"]) / 3 * x["volume"], axis=1
    )
    minute_data["apv"] = minute_data["pv"][back_time_index:].cumsum()
    minute_data["av"] = minute_data["volume"][back_time_index:].cumsum()

    minute_data["average"] = minute_data["apv"] / minute_data["av"]
    minute_data["vwap"] = minute_data.apply(
        lambda x: (x["close"] + x["high"] + x["low"]) / 3, axis=1
    )

    # print(f"\n{tabulate(minute_data, headers='keys', tablefmt='psql')}")
    if debug:
        tlog(
            f"\n{tabulate(minute_data[-110:-100], headers='keys', tablefmt='psql')}"
        )
        tlog(
            f"\n{tabulate(minute_data[-10:], headers='keys', tablefmt='psql')}"
        )

    return True
